"""Doc of bioat.

"""
from __future__ import annotations, absolute_import
from bioat.about import about
from bioat.bam import Bam
from bioat.fastx import Fastx
from bioat.hic import HiC
from bioat.logger import set_logging_level
from bioat.mgi import Mgi
from bioat.system import System
from bioat.table import Table
from bioat.version import (__version__, __author__, __doc_format__)

__name__ = "bioat"



"""
FASTQ 读取
FASTA 读取
...

load_fastx
load_fastx_generator

"""
import gzip
import sys
from Bio import SeqIO
from tqdm import tqdm

class Fastx():
    def __init__(self, file=None, iterable=True):
        self.file = file
        self.iterable = iterable
        self.fastx = None

    # def __load_fastx_list(self) -> list:
    #     """
    #
    #     :param str file: path of input <fastq | fastq.gz | fastx | fastx.gz>
    #     :return: a list for all reads:
    #         i.e. [['header', 'seq', 'info', 'quality'], [...],[...]]
    #     :rtype: list
    #     """
    #     f = open(self.file, 'rt') if not '.gz' in self.file else gzip.open(self.file, 'rt')
    #     # FASTQ @  FASTA >
    #     symbol = f.read(1)
    #     # print(symbol)
    #     f.close()
    #     # 全部载入内存就好，当文件很小的时候
    #     f = open(self.file, 'rt') if not '.gz' in self.file else gzip.open(self.file, 'rt')
    #
    #     if symbol == '@':
    #         # FASTQ
    #         # print('is fastq')
    #         ls = []
    #         read = []
    #         raw_info = [i.rstrip() for i in f.readlines()]
    #         line_fix = None  # 为了 fix quality 中开头为@的情况
    #
    #         # for line in tqdm(raw_info):
    #         for line in raw_info:
    #             if line.startswith('@'):
    #                 n = len(read)
    #                 if n == 0:
    #                     # 第一次循环开始
    #                     read.append(line)  # add header
    #                 elif n == 4:
    #                     # 后面的循环开始
    #                     ls.append(read)
    #                     read = []
    #                     read.append(line)  # add header
    #                 elif n == 3:
    #                     # try to add quality line
    #                     """Fix a bug.
    #                     如果第 4 行是@开头,先正常操作一行 read 的 append,再判断下一行是否是@开头,如果不是,文件损坏,如果是,没问题继续
    #                     ['@Beta12AdemL1C001R00100001768/1',
    #                     'ATCCCCGTATCTTCACCCCACCACAAACTATTAGCTTTAGA',
    #                     '+']
    #                     '@@IIEIBCE>IC<IBIIIIEAIEIEB<IDECCD6ICBCED<'
    #                     """
    #                     # print(read)
    #                     # print(line)
    #                     if line_fix:
    #                         # 遇到需要fix的下一次循环
    #                         read.append(line_fix)
    #                         ls.append(read)
    #                         read = []
    #                         read.append(line)
    #                         line_fix = None
    #                     else:
    #                         # 第一次遇到需要 fix 的 line
    #                         line_fix = line
    #                 else:
    #                     f.close()
    #                     raise ValueError('The file may be incomplete!')
    #             else:
    #                 read.append(line)  # add other three lines: seq, info, quality
    #
    #         ls.append(read)
    #         f.close()
    #         return ls
    #     elif symbol == '>':
    #         # FASTA
    #         # print('is fastx!')
    #         ls = []
    #         read = []
    #         seq = ''
    #         raw_info = [i.rstrip() for i in f.readlines()]
    #         # print(raw_info)
    #
    #         for line in raw_info:
    #             if line.startswith('>'):
    #                 # 读取 header！
    #                 n = len(read)
    #
    #                 if n == 0:
    #                     # 第一次循环
    #                     read.append(line)
    #                 elif n == 1:
    #                     # 已经有一个 header 了！现在缺 seq
    #                     read.append(seq)  # add seq line
    #                     ls.append(read)
    #                     read = []  # 重置 read 这个 list
    #                     read.append(line)
    #                     seq = ''  # 重置 seq 这个 str
    #                 else:
    #                     f.close()
    #                     raise ValueError('The file may be incomplete!')
    #             else:
    #                 # 读取并添加 seq
    #                 seq += line
    #
    #         read.append(seq)
    #         ls.append(read)
    #         return ls
    #     else:
    #         raise ValueError('Input line one must starts with @ for FQ or > for FA!')
    #     # f.close()
    #
    # def __load_fastx_generator(self):
    #     """
    #
    #     :param str file: path of input <fastq | fastq.gz | fastx | fastx.gz>
    #     :return: a generator for all reads:
    #         i.e. print(next(obj)) -> ['header', 'seq', 'info', 'quality']
    #     :rtype: generator
    #     """
    #     f = open(self.file, 'rt') if not '.gz' in self.file else gzip.open(self.file, 'rt')
    #     # FASTQ @  FASTA >
    #     symbol = f.read(1)
    #     # print(symbol)
    #     f.close()
    #     # 全部载入内存就好，当文件很小的时候
    #     f = open(self.file, 'rt') if not '.gz' in self.file else gzip.open(self.file, 'rt')
    #
    #     if symbol == '@':
    #         # FASTQ
    #         # print('is fastq!')
    #         # ls = []
    #         read = []
    #         line = f.readline().rstrip()
    #
    #         while True:
    #             if not line:
    #                 break
    #             else:
    #                 if line.startswith('@'):
    #                     n = len(read)
    #                     if n == 0:
    #                         # read == []
    #                         # 第一次循环开始
    #                         read.append(line)  # add header!
    #                         line = f.readline().rstrip()
    #                     elif n == 4:
    #                         # read == ['header', 'seq', 'info', 'quality']
    #                         # ls.append(read)
    #                         yield read
    #                         read = []
    #                         read.append(line)
    #                         line = f.readline().rstrip()
    #                     elif n == 3:
    #                         # TODO
    #                         """
    #                         @@IIEIBCE>IC<IBIIIIEAIEIEB<IDECCD6 # line! 期望它是 header！现在它是 quality
    #                         [
    #                             '@Beta12AdemL1C001R00100001768/1',
    #                             'ATCCCCGTATCTTCACCCCACCACAAACTATTAG',
    #                             '+',
    #                         ]'@@IIEIBCE>IC<IBIIIIEAIEIEB<IDECCD6'
    #
    #                         """
    #                         read.append(line)
    #                         line = f.readline().rstrip()
    #
    #                         if not line.startswith('@'):
    #                             raise ValueError('The file may be incomplete!')
    #                     else:
    #                         f.close()
    #                         raise ValueError('The file may be incomplete!')
    #
    #                     # header line!
    #                     # read.append(line)  # add header !
    #                 else:
    #                     # not header line!
    #                     read.append(line)
    #                     line = f.readline().rstrip()
    #
    #         # ls.append(read)
    #         yield read
    #         f.close()
    #         # return ls
    #     elif symbol == '>':
    #         # FASTA
    #         # print('is fastx!')
    #         ls = []
    #         read = []
    #         seq = ''
    #         # raw_info = [i.rstrip() for i in f.readlines()]
    #         line = f.readline().rstrip()
    #         # print(raw_info)
    #
    #         while True:
    #             if not line:
    #                 break
    #             else:
    #                 if line.startswith('>'):
    #                     # 读取 header！
    #                     n = len(read)
    #
    #                     if n == 0:
    #                         # 第一次循环
    #                         read.append(line)
    #                         line = f.readline().rstrip()
    #                     elif n == 1:
    #                         # 已经有一个 header 了！现在缺 seq
    #                         read.append(seq)  # add seq line
    #                         # ls.append(read)
    #                         yield read
    #                         read = []  # 重置 read 这个 list
    #                         read.append(line)
    #                         line = f.readline().rstrip()
    #                         seq = ''  # 重置 seq 这个 str
    #                     else:
    #                         f.close()
    #                         raise ValueError('The file may be incomplete!')
    #                 else:
    #                     # 读取并添加 seq
    #                     seq += line
    #                     line = f.readline().rstrip()
    #
    #         read.append(seq)
    #         # ls.append(read)
    #         yield read
    #         # return ls
    #     else:
    #         f.close()
    #         raise ValueError('Input line one must starts with @ for FQ or > for FA!')
    #     f.close()
    #
    # def _load_fastx(self):
    #     if self.iterable:
    #         return self.__load_fastx_generator()
    #     else:
    #         return self.__load_fastx_list()

    def get_headers(self):
        # if self.fastx is None:
        self.fastx = self._load_fastx()

        return [i[0][1:] for i in self.fastx]

    def get_sequence(self):
        # if self.fastx is None:
        self.fastx = self._load_fastx()

        return [i[1] for i in self.fastx]

    def get_record_dict(self):
        # if self.fastx is None:
        self.fastx = self._load_fastx()

        return {i[0][1:]: i[1] for i in self.fastx}

    # cli subcmd
    def filter_read_contains_N(self, file: str, output=sys.stdout.name):
        """Filter read contains N base in FASTA or FASTQ



        :param file: FASTA | FASTQ file name.
        :param output: FASTA | FASTQ file name file, stdout if not assigned, format as input.
        """
        if self.fastx is None:
            self.fastx = self._load_fastx()

        self.file = file
        output = sys.stdout if output == sys.stdout.name else output

        if isinstance(output, str):
            f = gzip.open(output, 'wt') if output.endswith('.gz') else open(output, 'wt')
        else:
            f = output

        fx = self._load_fastx()

        for read in fx:
            seq = read[1]

            if seq.upper().__contains__('N'):
                continue
            else:
                f.write('\n'.join(read) + '\n')
        f.close()


if __name__ == '__main__':
    # fastx = Fastx(file='../../data/random_l-180_n-100_hg38_minus-test.fa')
    pass



from __future__ import absolute_import

import time
import pandas as pd
from bioat.logger import set_logging_level
from pandarallel import pandarallel
import logging
from bioat._dev_tools import profile
import gc


try:
    pandarallel.initialize(
        progress_bar=False,
        use_memory_fs=True,
        verbose=1
    )
except SystemError:
    pandarallel.initialize(
        progress_bar=False,
        use_memory_fs=False,
        verbose=1
    )


class HiC:
    def __init__(self):
        pass

    @profile
    def get_effective_resolutions(self, genome_index, valid_pairs, output):
        """Get deepest resolution of cis-interation.

        输入为result文件，输出为matrix的大致resolution.

        :param genome_index: genome.fa.fai
        :param valid_pairs: rm_dup_pairs.allValidPairs by HiC-Pro
        :param output: table_output, TSV file
        """
        logging.debug('Develop mode is on')
        # load ref genome lengths
        df_chromosome = pd.read_csv(
            genome_index,
            sep='\t',
            usecols=[0, 1],
            names=['chromosome', 'length'],
            index_col='chromosome',
        )
        # load sample valid pairs by HiC-Pro after remove duplication and filter.
        reader = pd.read_csv(
            valid_pairs,
            sep='\t',
            header=None,
            names=[
                'read_name', 'chr_A', 'start_A', 'strand_A', 'chr_B', 'start_B', 'strand_B', 'insert_size',
                'fragment_name_A', 'fragment_name_B', 'mapQ_A', 'mapQ_B', 'allele_specific_info'
            ],
            usecols=['chr_A', 'start_A', 'chr_B', 'start_B'],
            dtype={
                'chr_A': str,
                'start_A': int,
                'chr_B': str,
                'start_B': int,
            },
            iterator=True
        )
        chunks = []

        while True:
            try:
                chunk = reader.get_chunk(10_000_000)
                chunks.append(chunk)
            except StopIteration:
                logging.debug("pandas reader iteration is done.")
                break

        df_valid_pairs = pd.concat(chunks, ignore_index=True)
        logging.debug(df_valid_pairs.info(memory_usage='deep'))
        # 思路：二分法确认new_bin下线
        output = open(output, 'wt')
        write_lines = 0
        output.write('test_bin(kb)\ttotal_bins\tthreshold(%)\tis_effective\n')
        write_lines += 1
        # 思路：二分法确认new_bin下线
        circle = 1
        new_bin = 512_000  # 512kb

        def get_bin_index(x):
            cumsum = df_chromosome['bins_cumsum'][x['chr_A']]
            bin_count_in_this_chrom_a = int(x['start_A'] / new_bin) + 1
            bin_count_in_this_chrom_b = int(x['start_B'] / new_bin) + 1
            bin_index_a = cumsum + bin_count_in_this_chrom_a - 1
            bin_index_b = cumsum + bin_count_in_this_chrom_b - 1
            return bin_index_a, bin_index_b

        while True:
            # 每次迭代都把new_bin / 2
            new_bin = int(new_bin / circle)
            circle = 2

            # calculate total bins
            df_chromosome['bins'] = df_chromosome['length'].map(lambda x: int(x / new_bin) + 1)
            df_chromosome['bins_cumsum'] = df_chromosome['bins'].cumsum()
            total_bins = df_chromosome.iloc[-1, 2]

            # calculate mapped bin index
            df_valid_pairs[['bin_index_A', 'bin_index_B']] = df_valid_pairs.parallel_apply(
            # df_valid_pairs[['bin_index_A', 'bin_index_B']] = df_valid_pairs.apply(
                get_bin_index,
                axis=1,
                result_type='expand'
            )

            # print(df_valid_pairs)
            # threshold calculation
            s_bin_idx_count1 = df_valid_pairs['bin_index_A'].value_counts()
            s_bin_idx_count2 = df_valid_pairs['bin_index_B'].value_counts()
            df_bin_interaction_count = pd.concat([s_bin_idx_count1, s_bin_idx_count2], axis=1)
            df_bin_interaction_count.fillna(0, inplace=True)
            df_bin_interaction_count['total'] = df_bin_interaction_count['bin_index_A'] + df_bin_interaction_count[
                'bin_index_B'
            ]
            threshold = round((df_bin_interaction_count['total'] >= 1000).sum() / total_bins * 100, 1)

            if threshold > 80:
                # 输出结果为有效,进入下一个循环
                output.write(f'{int(new_bin / 1000)}\t{total_bins}\t{threshold}\tYES\n')
                write_lines += 1
            else:
                # if write_lines != 1:
                #     # 第1+n次计算无效,直接终止程序
                #     output.write(f'{int(new_bin / 1000)}\t{total_bins}\t{threshold}\tNO\n')
                # else:
                #     # 第一次计算就无效,直接终止程序
                output.write(f'{int(new_bin / 1000)}\t{total_bins}\t{threshold}\tNO\n')
                break
            del s_bin_idx_count1, s_bin_idx_count2, df_bin_interaction_count
            gc.collect()
        output.close()


if __name__ == '__main__':
    set_logging_level(level='DEBUG')
    a = time.time()
    hic = HiC()
    hic.get_effective_resolutions(
        genome_index="/Volumes/zhaohn_HD/Bio/1.database/db_genomes/genome_fa/genome_ucsc_hg38/genome_ucsc_hg38.fa.fai",
        valid_pairs="../../data/hic/test.rm_dup_pairs.allValidPairs",
        output="/Users/zhaohuanan/Downloads/testout"
    )
    a = time.time() - a
    # print(f'\n{a:.3f}s')



import logging
from logging import (CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET)


def set_logging_level(level='ERROR'):

    level = level.upper()

    dt_level = {
        k: v
        for k, v in zip(
            ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'],
            [CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET]
        )
    }
    logging.basicConfig(level=dt_level[level])

    if level not in ('NOTSET', 'INFO'):
        logging.info(f'set logging level = {level}')


if __name__ == '__main__':
    #  ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'],
    # 一个比一个轻
    set_logging_level('CRITICAL')  # 只打印 CRITICAL
    # set_logging_level('ERROR')  # DEBUG、INFO、WARNING 没打印
    # set_logging_level('WARNING')  # DEBUG、INFO 没打印
    # set_logging_level('INFO')  # DEBUG 没打印
    # set_logging_level('DEBUG')  # 全打印
    # set_logging_level('NOTSET')  # 全打印

    logging.debug('Python debug')
    logging.info('Python info')
    logging.warning('Python warning')
    logging.error('Python Error')
    logging.critical('Python critical')




import sys
import pandas as pd


class Table():
    def __init__(self):
        pass

    def merge(
            self,
            inputs,
            tags,
            output,
            input_fmt='tsv',
            output_fmt='tsv',
            input_header=False,
            output_header=False
    ):
        """A simple tool to merge same formatted tables from different sample.

        Params
        :param inputs: table files
        :param tags: tags for each table
        :param output: merged file
        :param input_fmt: tsv | csv
        :param output_fmt: tsv | csv
        :param input_header: True | False, input table has header or not
        :param output_header: True | False, output table has header or not
        """
        # fix params
        inputs = list(inputs) if isinstance(inputs, tuple) else inputs.split(',')
        tags = list(tags) if isinstance(tags, tuple) else tags.split(',')
        input_header = 0 if input_header else None
        output = sys.stdout if output == sys.stdout.name else output
        dt_sep = {'csv': ',', 'tsv': '\t'}

        df = pd.DataFrame()

        for file, tag in zip(inputs, tags):
            _df = pd.read_csv(
                file,
                sep=dt_sep[input_fmt],
                header=input_header,
                index_col=None)
            _df.insert(0, column='<sample>', value=tag)
            df = pd.concat([df, _df], axis=0, ignore_index=True)

        df.to_csv(output, sep=dt_sep[output_fmt], header=output_header, index=None)


if __name__ == "__main__":
    pass



from __future__ import absolute_import
import fire
from bioat import about, Bam, Fastx, HiC, Mgi, System, Table, __version__


class Cli(object):
    """Cli interface of python package <bioat>

    - <bioat> is a commandline tool and a python package, which is short for
        Bioinformatic Analysis Tools.
    - <bioat> has many subcommand to deal with different bio-format:
        BED, BAM, FASTA, FASTQ, VCF, et al.
    """

    def __init__(self):
        self.bam = Bam()
        # self.fastx = Fastx()
        self.hic = HiC()
        self.mgi = Mgi()
        # self.system = System()
        self.table = Table()

    def about(self):
        """Print information about <bioat>.
        """
        return about

    def list(self):
        """Print commands and subcommands.
        """
        for att in dir(self):
            if not att.startswith('_'):
                print(f'{att}')

                for sub_att in dir(getattr(self, att)):
                    if not sub_att.startswith('_'):
                        print(f'  ├── {sub_att}')

    def version(self):
        """Print version information.
        """
        return __version__


def main() -> int:
    calculator = Cli()
    fire.Fire(calculator, name='bioat')


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from bioat.version import (__version__, __author__, __upgrade_date__, __name__, __email__, __doc_format__)

about = f"""
BioinformaticAnalysisTools ({__name__})
    - bioat version:
        {__version__}
    - repository address:
        https://github.com/hermanzhaozzzz/BioinformaticAnalysisTools

    - author:
        - name: {__author__}
        - email: {__email__}
    - doc:
        - format: {__doc_format__}
    - last update:
        - {__upgrade_date__}
"""
